#!/usr/bin/env python
"""
loaders

Util methods to load things on the fly, optionally type checking them
if needed.

"""

import importlib
from .errors import LoaderError


def parse_module_class_name(module_dot_class):
    """
    _parse_module_class_name_

    parse a string of the form module.submodule.Class

    :param module_dot_class: String containing module.Class style ref

    :returns: the modulename and the class name as a tuple (mod, cls)

    """
    try:
        split_name = module_dot_class.rsplit(".", 1)
    except IndexError:
        # string not contanining enough dots
        msg = "Couldn't split mod.class name: {0}".format(module_dot_class)
        raise LoaderError(msg, plugin=module_dot_class)
    except AttributeError:
        msg = "Plugin name is not a string: {0}".format(module_dot_class)
        raise LoaderError(msg, plugin=module_dot_class)
    if len(split_name) != 2:
        msg = (
            "Expected two elements of mod.class string in {0}"
            " which resulted in {1}"
            ).format(module_dot_class, split_name)
        raise LoaderError(msg, plugin=module_dot_class)
    return split_name


def load_object(plugin_name):
    """
    _load_object_

    Given a module.submodule.thing style string,
    load thing and return it.
    Assumes that the last element is an attr in the
    module denoted by the previous elements

    :param plugin_name: string, name of plugin of form module.submod.thing

    :returns: Imported reference to thing.

    """
    modname, objname = parse_module_class_name(plugin_name)

    try:
        modref = importlib.import_module(modname)
    except ImportError as ex:
        msg = "Error Importing Module: {0}\n".format(modname)
        msg += str(ex)
        raise LoaderError(msg, plugin=plugin_name)
    objref = getattr(modref, objname, None)
    if objref is None:
        msg = "Object {0} is not defined in Module {1}\n".format(
            objname, modname)
        msg += "Unable to load plugin"
        raise LoaderError(msg, plugin=plugin_name)
    return objref


def load_plugin(plugin_name, type_check=None):
    """
    _load_plugin_

    Given a plugin name of the form mod.submod.Class, load the
    module and return an instance of the class.

    To verify the plugin is of the appropriate type, pass a class
    ref as type_check to perform an issubclass test on the
    loaded class

    :param plugin_name: string, name of plugin of form module.Class
    :param type_check: Class object or tuple of objects to be passed to
       issubclass to verify type of plugin

    :returns: Class Reference for the plugin

    """
    classref = load_object(plugin_name)
    if type_check is not None:
        if not issubclass(classref, type_check):
            msg = "Plugin in {0} is not a {1} subclass:\n".format(
                plugin_name, type_check.__name__
            )
            raise LoaderError(msg, plugin=plugin_name)
    return classref



#!/usr/bin/env python
"""
_factory_plugin_

Plugin base class that can be inherited to register
a class with a factory

"""
import six

from .registry import get_factory
from .registry import PluggageMeta

@six.add_metaclass(PluggageMeta)
class PluggagePlugin(object):
    """
    _PluggagePlugin_

    Inherit from this class to trigger registration
    with the pluggage.registry factory.

    Override the PLUGGAGE_FACTORY_NAME to be the
    name of the factory to register with, and
    the PLUGGAGE_OBJECT_NAME can be used to set the
    name of the plugin, if not overridden (or None)
    the name of the class will be used.

    Eg:

    SomeClass(PluggagePlugin):
        PLUGGAGE_FACTORY_NAME = 'my_factory'

    Will register the plugin as SomeClass under
    my_factory leading to instantiation patterns like:
    f = get_factory('my_factory')
    some_class_instance = f('SomeClass')

    """
    PLUGGAGE_FACTORY_NAME = 'abstract'
    PLUGGAGE_OBJECT_NAME = None

    def __init__(self):
        super(PluggagePlugin, self).__init__()
        self._factory_ref = None

    def _get_factory(self):
        """
        _get_factory_

        Helper method to get the appropriate factory
        for this plugin

        """
        if self._factory_ref is None:
            self._factory_ref = get_factory(self.PLUGGAGE_FACTORY_NAME)
        return self._factory_ref



#!/usr/bin/env python
"""
_registry_

Plugin class registry utils

"""
import pkg_resources
import importlib

from .errors import FactoryError


class Registry(object):
    """
    _Registry_

    Plugin registry. Wrapper for a plugin registry.
    Instantiate with the name of the factory and then use
    it to access plugins registered with that factory.

    The __call__ operator will instantiate a named plugin
    passing optional args/kwargs to the named plugin class

    """
    # singleton registry of factory name: plugin name:cls mapping
    _REGISTRY = {}

    def __init__(self, factory_name):
        self.factory_name = factory_name
        self.registry = self._REGISTRY.setdefault(factory_name, {})

    def register(self, obj_name, cls):
        if obj_name in self.registry:
            msg = (
                "Duplicate Plugin registered for {} "
                "register: {}, existing {}"
            ).format(
                obj_name,
                cls.__name__,
                self.registry[obj_name].__name__
            )
            raise FactoryError(
                msg, factory=self.factory_name, plugin=obj_name
            )
        self.registry[obj_name] = cls

    def get(self, plugin, default=None):
        """
        get with default access to test if plugins are
        registered or just get access to the class object

        :param plugin: Name of the plugin
        :param default: optional, thing to return if
           not present
        """
        return self.registry.get(plugin, default)

    def __call__(self, plugin, *args, **kwargs):
        cls = self.get(plugin)
        if cls is None:
            msg = (
                "No plugin registered with {} for name {}"
            ).format(self.factory_name, plugin)
            raise FactoryError(
                msg, factory=self.factory_name,  plugin=plugin
            )
        return cls(*args, **kwargs)


def get_factory(factory_name, throw_on_nonexist=False, load_modules=None):
    """
    get an instance of the Registry object for the named
    factory. Note that this will create a new
    empty entry in the registry for factory name if it
    doesnt exist, unless the throw_on_nonexist is True

    :param factory_name: Name of the factory to get plugins for
    :param throw_on_nonexist: If no plugins exist for the factory,
        raise an exception
    :param load_modules: List of module names to import to trigger
        plugin registration

    :returns: Instance of Registry set up to use the given factory
        name
    """
    if load_modules is not None:
        for module_name in load_modules:
            importlib.import_module(module_name)

    if (factory_name not in Registry._REGISTRY) and throw_on_nonexist:
        msg = "Factory name {} not found in registry".format(factory_name)
        raise FactoryError(msg, factory=factory_name)
    return Registry(factory_name)


def init_factory(factory_name, throw_on_nonexist=False, load_modules=None):
    """
    Similar to get_factory but uses the pluggage_modules
    entry points to load plugin modules registered by other packages.
    Call this from a main routine to initialise a factory
    with pluggage modules entry points defined in setuptools

    Arguments are same as for get_factory

    :returns: Instance of Registry set up to use the given factory
        name

    """
    for entry_point in pkg_resources.iter_entry_points(group='pluggage_modules'):
        entry_point.load()
    return get_factory(factory_name, throw_on_nonexist=False, load_modules=None)


class PluggageMeta(type):
    """
    metaclass for Pluggage derived objects that register the
    objects with the appropriate registry/factory object
    """
    def __init__(cls, name, bases, dct):
        factory = get_factory(cls.PLUGGAGE_FACTORY_NAME)
        object_name = cls.PLUGGAGE_OBJECT_NAME
        if object_name is None:
            object_name = cls.__name__
        factory.register(object_name, cls)
        super(PluggageMeta, cls).__init__(name, bases, dct)



#!/usr/bin/env python
"""
plugins

Simple factory/plugin loader that implements a dictionary
interface around a set of plugins and
imports/loads them on the fly via the dictionary key access
API

"""

from .loaders import load_object
from .errors import LoaderError


class Plugins(dict):
    """
    _Plugins_

    Dictionary based object that provides simple
    object loading on get/getitem and a couple
    of helpers to instantiate or call plugins

    """
    def __init__(self, **plugins):
        super(Plugins, self).__init__()
        self.update(plugins)

    def __getitem__(self, key):
        """
        Implement [key] access to load the
        plugin object and return it
        """
        if key in self:
            return super(Plugins, self).__getitem__(key)
        try:
            obj_ref = load_object(key)
        except LoaderError:
            raise KeyError(key)
        self[key] = obj_ref
        return obj_ref

    def get(self, key, default=None):
        """
        _get_

        Implement get call to return a plugin or
        default if it doesnt exist
        """
        if key not in self:
            return default
        return self[key]

    def __call__(self, name, *args, **kwargs):
        """
        _call_

        Get the named plugin and return the results of
        calling it with args/kwargs passed through to it.

        """
        plugin = self[name]
        if not callable(plugin):
            raise RuntimeError(
                "Plugin {0} from {1} is not callable".format(
                    name, super(Plugins, self).get(name)
                )
            )

        return plugin(*args, **kwargs)



#!/usr/bin/env python
"""
_errors_

Exception classes for the package

"""


class PluggageError(Exception):
    """
    base class for package errors

    """
    pass


class LoaderError(PluggageError):
    """
    LoaderError

    Indicates a failure to dynamically load a plugin
    """
    def __init__(self, message, plugin=None):
        super(LoaderError, self).__init__(message)
        self.plugin = plugin


class FactoryError(PluggageError):
    """
    FactoryError

    Indicates that there is an error with a factory
    plugin or registry error

    """
    def __init__(self, message, factory=None, plugin=None):
        super(FactoryError, self).__init__(message)
        self.plugin = plugin
        self.factory = factory



#!/usr/bin/env python

from setuptools import setup

setup(name='pluggage',
      version='0.0.4',
      description='Plugin, dynamic loading and factory utility library ',
      author='Dave Evans',
      author_email='evansde77.github@gmail.com',
      url='https://pluggage.readthedocs.org/en/latest/',
      packages=['pluggage'],
      package_dir={'pluggage': 'src/pluggage'},
      entry_points = { 
        'pluggage_modules': [
         ]
       }
     )
